package polyglot.ext.ejava;

import java.io.Reader;
import java.util.HashMap;
import java.util.List;

import polyglot.ast.NodeFactory;
import polyglot.ext.ejava.ast.EJavaNodeFactory_c;
import polyglot.ext.ejava.parse.Grm;
import polyglot.ext.ejava.parse.Lexer_c;
import polyglot.ext.ejava.types.EJavaTypeSystem;
import polyglot.ext.ejava.types.EJavaTypeSystem_c;
import polyglot.ext.ejava.visit.*;
import polyglot.frontend.*;
import polyglot.lex.Lexer;
import polyglot.types.TypeSystem;
import polyglot.util.ErrorQueue;

/**
 * Extension information for ejava extension.
 */
public class ExtensionInfo extends polyglot.ext.jl5.ExtensionInfo {
    static {
        // force Topics to load
        Topics t = new Topics();
    }

    public String defaultFileExtension() {
        return "ejava";
    }

    public String compilerName() {
        return "ejavac";
    }

    public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        Lexer lexer = new Lexer_c(reader, source.name(), eq);
        Grm grm = new Grm(lexer, ts, nf, eq);
        return new CupParser(grm, source, eq);
    }

    protected NodeFactory createNodeFactory() {
        return new EJavaNodeFactory_c();
    }

    protected TypeSystem createTypeSystem() {
        return new EJavaTypeSystem_c();
    }

    public static final Pass.ID TRANSFORM_EXTENSION_BODY = new Pass.ID("transform-extension-body");
    public static final Pass.ID PRE_GATHER_TOPLEVEL_EXTENSIONS = new Pass.ID("pre-gather-toplevel-extensions");
    public static final Pass.ID GATHER_TOPLEVEL_EXTENSIONS = new Pass.ID("gather-toplevel-extensions");
    public static final Pass.ID PRE_CONSOLIDATE_EXTENSIONS = new Pass.ID("pre-consolidate-extensions");
    public static final Pass.ID CONSOLIDATE_EXTENSIONS = new Pass.ID("consolidate-extensions");
    public static final Pass.ID CONSOLIDATE_EXTENSIONS_ALL = new Pass.ID("consolidate-extensions-all");
    public static final Pass.ID CLEAR_EXTENSION_CASES = new Pass.ID("clear-extension-cases");
    public static final Pass.ID ENFORCE_EXTENSION_BODY_RESTRICTIONS = new Pass.ID("enforce-extension-body-restrictions");
    public static final Pass.ID EXTENSION_APPLICATOR = new Pass.ID("apply-extensions");

    public List passes(Job job) {
        List passes = super.passes(job);
        // TODO: add passes as needed by your compiler

        // replace the original type builder with one that knows about
        // extensions
        replacePass(passes, Pass.BUILD_TYPES, new VisitorPass(Pass.BUILD_TYPES, job, new EJavaTypeBuilder(job, (EJavaTypeSystem) ts, nf)));

        //replace the original pass with one that will keep current context with type system
        replacePass(passes, Pass.TYPE_CHECK, new VisitorPass(Pass.TYPE_CHECK, job, new EJavaTypeChecker(job, ts, nf)));

        // complain if extension body is accessing protected methods of
        // "base" class that will result in javac compilation errors
        beforePass(passes, Pass.PRE_OUTPUT_ALL, 
                new VisitorPass(ENFORCE_EXTENSION_BODY_RESTRICTIONS, job, new ExtensionBodyRestrictionEnforcer(job,ts, nf)));
        // prepare extension bodies for code generation
        afterPass(passes, ENFORCE_EXTENSION_BODY_RESTRICTIONS, new VisitorPass(TRANSFORM_EXTENSION_BODY, job, new ExtensionBodyTransformer(job, ts, nf)));

        HashMap extensions = new HashMap();

        // gather all top-level extensions, put them in the hashmap
        afterPass(passes, TRANSFORM_EXTENSION_BODY, 
                new VisitorPass(PRE_GATHER_TOPLEVEL_EXTENSIONS,job, new TopLevelExtensionGatherer(job, ts, nf, extensions)));

        // gather extension cases, make them part of the appropriate top-level
        // extension
        afterPass(passes, PRE_GATHER_TOPLEVEL_EXTENSIONS,
                new VisitorPass(PRE_CONSOLIDATE_EXTENSIONS, job, new ExtensionConsolidator(job, ts,nf, extensions)));

        afterPass(passes, PRE_CONSOLIDATE_EXTENSIONS, 
                new VisitorPass(EXTENSION_APPLICATOR, job, new ExtensionApplicator(job, ts, nf, extensions)));

        afterPass(passes, EXTENSION_APPLICATOR, 
                new VisitorPass(CLEAR_EXTENSION_CASES, job, new ExtensionCaseClearer(job, ts, nf)));

        extensions = new HashMap();

        // gather all top-level extensions, put them in the hashmap
        afterPass(passes, CLEAR_EXTENSION_CASES, 
                new VisitorPass(GATHER_TOPLEVEL_EXTENSIONS, job, new TopLevelExtensionGatherer(job, ts, nf, extensions)));

        // gather extension cases, make them part of the appropriate top-level
        // extension
        afterPass(passes, GATHER_TOPLEVEL_EXTENSIONS, 
                new VisitorPass(CONSOLIDATE_EXTENSIONS, job, new ExtensionConsolidator(job, ts, nf, extensions)));
        afterPass(passes, CONSOLIDATE_EXTENSIONS, new BarrierPass(CONSOLIDATE_EXTENSIONS_ALL, job));

        return passes;
    }

}
