package polyglot.ext.ejava.visit;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import polyglot.ast.*;
import polyglot.ext.ejava.ast.EJavaNodeFactory;
import polyglot.ext.ejava.ast.ExtensionApplication;
import polyglot.ext.ejava.ast.ExtensionDecl;
import polyglot.ext.ejava.ast.Unextended;
import polyglot.ext.ejava.types.EJavaContext;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.Position;
import polyglot.visit.AscriptionVisitor;
import polyglot.visit.NodeVisitor;

/*******************************************************************************
 * Visitor that applies extensions wherever needed so that the stufff done by
 * the extension type inference doesn't make javac go wrong.
 ******************************************************************************/
public class ExtensionApplicator extends AscriptionVisitor {
    protected HashMap extensions;

    public ExtensionApplicator(Job job, TypeSystem ts, NodeFactory nf, HashMap extensions) {
        super(job, ts, nf);
        this.extensions = extensions;
    }

    protected ExtensionType extension() {
        Type t = context().currentClass();
        if (t instanceof ExtensionType)
            return (ExtensionType) t;
        else
            return null;
    }

    protected boolean insideExtension() {
        return extension() != null;
    }

    protected ExtensionType findExtension(ClassType from, ClassType to, EJavaContext context) {
        // System.out.println("findextension("+from+", "+to+")");
        // System.out.println("ctxt="+context);
        List extensions = context.extensions();
        if (extensions == null)
            extensions = new LinkedList();
        for (Iterator ee = extensions.iterator(); ee.hasNext();) {
            ExtensionType et = (ExtensionType) ee.next();
            // System.out.println("trying " + et);
            if (!ts.isSubtype(from, et.superType()))
                continue;
            for (Iterator ii = et.interfaces().iterator(); ii.hasNext();) {
                Type i = (Type) ii.next();
                if (ts.isImplicitCastValid(i, to))
                    return et;
            }
        }
        context = (EJavaContext) context.pop();
        if (context == null)
            return null;
        else
            return findExtension(from, to, context);
    }

    
    public Expr ascribe(Expr e, Type toType) throws SemanticException {
        if (toType instanceof ClassType) {
            ClassType clToType = (ClassType) toType;
            // Don't apply other extensions to an expression with an explicit
            // extension application. This is probably not needed, but it
            // doesn't hurt.
            if (e instanceof Unextended)
                return e;
            if (e instanceof ExtensionApplication) {
                boolean augmented = false;
                ExtensionType et = (ExtensionType) ((ExtensionApplication) e).extension().type();
                String compilationUnit = e.position().file();
                if (context.package_() != null)
                    compilationUnit = context.package_() + "." + compilationUnit;
                HashMap extensionsInThisCU = (HashMap) extensions.get(compilationUnit);
                if (extensionsInThisCU != null
                        && ((ExtensionDecl) extensionsInThisCU.get(et.name())).isLocalTop())
                    augmented = true;
                ((ExtensionApplication) e).augmented(augmented);
                return e;
            }

            if ((!ts.isImplicitCastValid(e.type(), clToType)) && (!(e.type() instanceof PrimitiveType))) {
                ExtensionType et = clToType instanceof ExtensionType ? 
                        (ExtensionType) clToType : findExtension((ClassType) e.type(), clToType, (EJavaContext) context());
                if (et != null) {
                    boolean augmented = false;
                    String compilationUnit = e.position().file();
                    if (context.package_() != null)
                        compilationUnit = context.package_() + "." + compilationUnit;
                    HashMap extensionsInThisCU = (HashMap) extensions.get(compilationUnit);
                    if (extensionsInThisCU != null && ((ExtensionDecl) extensionsInThisCU.get(et.name())).isLocalTop())
                        augmented = true;

                    if (e.type() instanceof ExtensionType && ((ExtensionType) e.type()).extensionName().equals(et.name())) {
                    } else {
                        e = ((EJavaNodeFactory) nf).ExtensionApplication(
                                Position.COMPILER_GENERATED, e, nf.CanonicalTypeNode(Position.COMPILER_GENERATED, et), augmented);
                    }

                }
            }
        }
        return e;
    }

    public Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException {
        if (old instanceof Import) {
            Import imold = (Import) old;

            Node r = super.leaveCall(imold, n, v);
            try {
                Named nt = typeSystem().forName(imold.name());
                if (nt instanceof ExtensionType) {
                    // System.out.println("whooopeeee!");
                    ((EJavaContext) context()).addExtension((ExtensionType) nt);
                    // System.out.println("added extension, ctxt is
                    // "+context());
                }
            } catch (SemanticException e) {
            }
            return r;
        } else {

            Node r = super.leaveCall(old, n, v);
            // check for "stupid" extension applications (these must be
            // explicit)
            if (r instanceof ProcedureCall) {
                // first, complain about "stupid" extension applications in the
                // receiver
                if (r instanceof Call) {
                    Call c = (Call) r;
                    Receiver target = c.target();
                    Type targetType = target != null ? target.type() : null;
                    if (targetType != null && target instanceof ExtensionApplication) {
                        // this was an explicit "with", i.e. it wasn't
                        // automatically
                        // inserted by compiler. (we know this because the
                        // expression has a
                        // type - had it been put there by ascribe() above, it
                        // wouldn't have
                        // one.)
                        ExtensionApplication ea = (ExtensionApplication) target;
                        ExtensionType et = (ExtensionType) ea.extension().type();
                        if (!et.hasMethod(c.methodInstance()))
                            throw new SemanticException("stupid extension application");
                    }
                }
                // then, do the same for the arguments
                ProcedureCall pc = (ProcedureCall) r;
                ProcedureInstance pi = pc.procedureInstance();
                // (TODO)
            } else if (r instanceof ExtensionApplication) {
                // This is very much like the procedure call case above.
                ExtensionApplication outerEA = (ExtensionApplication) r;
                ExtensionType outerET = (ExtensionType) outerEA.extension().type();
                if (outerEA.expression() instanceof ExtensionApplication
                        && outerEA.expression().type() != null) {
                    // the inner extension application is explicit
                    ExtensionApplication innerEA = (ExtensionApplication) outerEA.expression();
                    ExtensionType innerET = (ExtensionType) innerEA.extension().type();
                    if (ts.isImplicitCastValid(innerEA.expression().type(), outerET.superType()))
                        throw new SemanticException("stupid extension application");
                }
            }
            return r;
        }

    }

}
